"""This module contains various relatively low-level classes to modify a PAM
configuration set"""

try: import logging
except ImportError: import compat.dummy_logging as logging

_theLogger = logging.getLogger("PAMConfig.modifiers")

import os.path

from PAMConfig.confset import ConfAtom

class _BaseConfModifier(object):
    def __init__(self, facility, module, modArgs, modFollowArgs,  deny = 'pam_deny.so', permit = 'pam_permit.so'):
        self._module = module
        self._modArgs = modArgs[:]
        self._modFollowArgs = modFollowArgs[:]
        self._deny = deny
        self._permit = permit
        self.facility = facility
    # enddef __init__

    def _incPrecedingJumps(self, conf, position, distance = 0):
        return self._modPrecedingJumps(conf, position, distance, 1)
    # enddef _incPrecedingJumps

    def _decPrecedingJumps(self, conf, position, distance = 0):
        return self._modPrecedingJumps(conf, position, distance, -1)
    # enddef _decPrecedingJumps

    def _modPrecedingJumps(self, conf, position, distance = 0, change = 1):
        # adjust previous jumps
        for prevpos in xrange(position, -1, -1):
            adjAtom = conf[prevpos][0]
##            logging.debug("position=%d, atom=%s" % (prevpos, adjAtom))
            if adjAtom.type == adjAtom.TYPE_CONF_INCLUDE:
                break # don't get into outer services if it was pam_stack style inclusion
            adjAtom.controlUpdateJumps(delta = change, filt = lambda (c, j): j >= distance)
            distance += 1
        # endfor
    # enddef _modPrecedingJumps

    def _nextIs(self, modname, conf, position):
        """Checks if the specified modname goes immediately after specified
        conf item (i.e. (atom, src, position))"""
        if len(conf) > position + 1:
            position += 1
        else:
            return False
        atom = conf[position][0]
        return atom.type == atom.TYPE_CONF and atom.module == modname
    # enddef _nextIs

    def remove(self, confSet, service, position): #, affected):
        """removes a module at the given position and change the stack
        appropriately"""
##        """removes a module at the given position and change affected
##        stack(s) appropriately"""
        conf = confSet.getConf(service, self.facility)
        (theAtom, src, listpos) = conf[position]
        src.atoms[listpos] = ConfAtom.CommentAtom(str(theAtom))
        self._decPrecedingJumps(conf, position - 1, 1)
        # FIXME: don't deal with affected services at all at the moment.
        # this is perfectly enough for RedHat and "recent derivatives"
        # but probably somewhat dangerous for other styles of inclusion.
        # But actually no known system has PAM jumps in "end user" services
        # (those which use one of common services) and system administrator
        # must be brain-dead to use this feature, because a PAM configuration
        # quickly becomes unmaintainable after such uses.
    # enddef remove
# endclass _BaseConfModifier
        
class AuthConfModifier(_BaseConfModifier):
    """This class allows to add a module into a specified place
    in the PAM atoms chain (typically read from a file)"""
    def __init__(self, module, modArgs = [], modFollowArgs = ['try_first_pass'], deny = 'pam_deny.so', permit = 'pam_permit.so'):
        super(AuthConfModifier, self).__init__('auth', module,
                                               modArgs, modFollowArgs,
                                               deny, permit)
    # enddef __init__

    def addToRequired(self, confSet, service, position):
        """adds the module after a required module which doesn't
        have any subsequent modules (the "no tail" case)"""

        conf = confSet.getConf(service, self.facility)
        location = conf[position]
        (theAtom, src, i) = location
        
        assert theAtom.type == theAtom.TYPE_CONF

        # since the atom control is changed to sufficient
        theAtom.controlSet('sufficient')
        tgtAtom = confSet.createConfAtom(service, self.facility,
                                         'required', self._module,
                                         self._modArgs + self._modFollowArgs)
        src.atoms.insert(i+1, tgtAtom)
        # actually we don't even need to add a deny in this case

##        if not self._nextIs(self._deny, confSet, service, (theAtom, src, i)):
##            deny = confSet.createConfAtom(service, self.facility,
##                                          'required', self._deny)
##            src.atoms.insert(i+2, deny)
        return None

    def addToRequiredTailed(self, confSet, service, position):
        """adds the module after a required module which doesn't
        have any subsequent modules (the "has tail" case)"""
        conf = confSet.getConf(service, self.facility)
        location = conf[position]
        (theAtom, src, listpos) = location
        
        assert theAtom.type == theAtom.TYPE_CONF

        tgtAtom = confSet.createConfAtom(service, self.facility,
                                         theAtom.control, self._module,
                                         self._modArgs + self._modFollowArgs)
        src.atoms.insert(listpos + 1, tgtAtom)

        # add pam_permit.so if it's not already there
        if not self._nextIs(self._permit, conf, position):
            perm = confSet.createConfAtom(service, self.facility,
                                          'required', self._permit)
            src.atoms.insert(listpos + 2, perm)
        # endif

        # modify the atom flag (make errors non-fatal and 
        theAtom.controlSet('sufficient')
        theAtom.controlUpdate({'success': 'ok', 'new_authtok_reqd': 'ok'})

        self._incPrecedingJumps(conf, position)
    # enddef addToRequiredTailed

    def addToSufficient(self, confSet, service, position):
        """adds the module after a "sufficient" module"""
        conf = confSet.getConf(service, self.facility)
        location = conf[position]
        (theAtom, src, listpos) = location
        
        assert theAtom.type == theAtom.TYPE_CONF

        tgtAtom = confSet.createConfAtom(service, self.facility,
                                         theAtom.control, self._module,
                                         self._modArgs + self._modFollowArgs)
        src.atoms.insert(listpos + 1, tgtAtom)

        self._incPrecedingJumps(conf, position)
    # enddef addToSufficient

    def removeRequired(self, confSet, service, position, affected):
        """removes 'auth' module which is 'required' or 'required-like'"""
        conf = confSet.getConf(service, self.facility)
        location = conf[position]
        (theAtom, src, listpos) = location
        assert position > 0

        conf[position - 1][0].controlSet(theAtom.control)
        src.atoms[listpos] = ConfAtom.CommentAtom(str(theAtom))
##        del src.atoms[listpos]
        self._decPrecedingJumps(conf, position - 2, 2)
    # enddef removeRequired

    def removeRequiredSingle(self, confSet, service, position, affected):
        conf = confSet.getConf(service, self.facility)
        location = conf[position]
        (theAtom, src, listpos) = location

        src.atoms[listpos] = ConfAtom.CommentAtom(str(theAtom))
        deny = confSet.createConfAtom(service, self.facility,
                                      'required', self._deny)
        src.atoms.insert(listpos + 1, deny)
    # enddef removeRequiredSingle

    def removeSufficient(self, confSet, service, position, affected):
        conf = confSet.getConf(service, self.facility)
        location = conf[position]
        (theAtom, src, listpos) = location

        src.atoms[listpos] = ConfAtom.CommentAtom(str(theAtom))
        self._decPrecedingJumps(conf, position - 1, 1)
    # enddef removeSufficient
# endclass AuthConfModifier

def addService(service, facility, confSet, module, modArgs = []):
    svc = confSet.addService(service)
    svc.append(confSet.createConfAtom(service, facility, "required", module, modArgs))
    return svc
# enddef addService
    

class PasswordConfModifier(_BaseConfModifier):
    def __init__(self, module, modArgs = [], modFollowArgs = ['try_first_pass'], deny = 'pam_deny.so', permit = 'pam_permit.so'):
        super(PasswordConfModifier, self).__init__('password', module,
                                                   modArgs, modFollowArgs,
                                                   deny, permit)
    # enddef __init__

    def addToRequired(self, confSet, service, position):
        conf = confSet.getConf(service, self.facility)
        location = conf[position]
        (theAtom, src, listpos) = location
        
        assert theAtom.type == theAtom.TYPE_CONF

        tgtAtom = confSet.createConfAtom(service, self.facility,
                                         'optional', self._module,
                                         self._modArgs + self._modFollowArgs)
        src.atoms.insert(listpos + 1, tgtAtom)

        # FIXME: do we really need to adjust these jumps?
        self._incPrecedingJumps(conf, position - 1, 1)
    # enddef addToRequired

    def addToSufficient(self, confSet, service, position):
        conf = confSet.getConf(service, self.facility)
        (theAtom, src, listpos) = conf[position]
        
        assert theAtom.type == theAtom.TYPE_CONF

        tgtAtom = confSet.createConfAtom(service, self.facility,
                                         'optional', self._module,
                                         self._modArgs + self._modFollowArgs)
        src.atoms.insert(listpos, tgtAtom)
##        theAtom.controlSet('required')

        # FIXME: do we really need to adjust these jumps?
        self._incPrecedingJumps(conf, position - 1, 1)
    # enddef addToSufficient

    def removeRequiredSingle(self, confSet, service, position):
        conf = confSet.getConf(service, self.facility)
        (theAtom, src, listpos) = conf[position]
        src.atoms[listpos] = ConfAtom.CommentAtom(str(theAtom))
        deny = confSet.createConfAtom(service, self.facility,
                                      'required', self._deny)
        src.atoms.insert(listpos + 1, deny)
    # enddef removeRequiredSingle
# endclass PasswordConfModifier
        
        
