# -*- coding: utf-8; mode: python -*-

import sys
import warnings

import string
import os.path
import os
import errno
import random

try: import logging
except ImportError: import compat.dummy_logging as logging

# Evil _ancient_ python workaround ;-)
try:
    False
except NameError:
    False = 0
    True = 1

_theLogger = logging.getLogger('PAMConfig.support')


def testset(predicate, it):
    """Test if the predicate is True for at least one element in a given set

    @param predicate: any predicate which accepts one argument
    @param it: an iterable
    @rtype: bool
    """
    for el in iter(it):
        if predicate(el):
            return True
    return False
# enddef testset

class _RndPartGen(object):
    """Random part name generator"""

    # For OSes with case-insensitive filenames
    # choose only appropriate letters
    if os.path.normcase(string.ascii_lowercase) == os.path.normcase(string.ascii_uppercase):
        VALID_SYMBOLS = os.path.normcase(string.ascii_lowercase) + string.digits + "!@#_-"
    else:
        VALID_SYMBOLS = string.ascii_letters + string.digits + "!@#_-"

    PART_LEN = 6

    def __init__(self, rnd = random.Random()):
        """creates a new _RndPartGen instance. Optional
        rnd argument allows to specify a custom random generator
        (including thread-safe ones if appropriate), default is
        random.Random()
        """
        self._rnd = rnd
    # enddef __init__

    def __iter__(self):
        return self

    def next(self):
        choice = self._rnd.choice
        return ''.join([choice(self.VALID_SYMBOLS) \
                        for _ in xrange(self.PART_LEN)])

    # enddef next
# endclass _RndPartGen

class _SafeNewFile(object):
    MAX_ATTEMPTS = sys.maxint # heh-heh :-)
    RND_GEN = _RndPartGen()
    def __init__(self, fname, prefix = ".", suffix = ".", mode=0644):

        self.name = None
        self.file = None

        flag = os.O_CREAT | os.O_WRONLY | os.O_EXCL
        try:
            fd = os.open(fname, flag, mode)
        except OSError, e:
            if e[0] != errno.EEXIST:
                raise # re-raise the exception
        else:
            self.name = fname
            self.file = os.fdopen(fd, 'wb')
            return
        # have to try a randomed name
        (dname, fname) = os.path.split(fname)
        for _ in xrange(self.MAX_ATTEMPTS):
            rndpart = self.RND_GEN.next()
            fullname = os.path.join(dname, prefix + fname + suffix + rndpart)
            try:
                fd = os.open(fullname, flag, mode)
            except OSError, e:
                if e[0] != errno.EEXIST:
                    raise # re-raise the exception
            else:
                self.name = fullname
                self.file = os.fdopen(fd, 'wb')
                return
        # emm? really?
        raise OSError(errno.EBUSY,
                      "unable to create an unique file after %d tries" % self.MAX_ATTEMPTS)
    # enddef __init__

    def __getattr__(self, attr):
        # lazy redirector to self.file
        v = getattr(self.file, attr)
        setattr(self, attr, v)
        return v
    # enddef __getattr__
# endclass _SafeNewFile

class _NewConfigFile(_SafeNewFile):
    def safe_unlink(self):
        if self.name:
            try:
                os.unlink(self.name)
            except OSError:
                pass
    # enddef unlink

    def moveToPlace(self, new_location):
        """moves itself into a permanent location.
        raises appropriate OSError if target location
        is occupied by another file. Note that
        upon successful completion the function shouldn't
        be called again
        """
        assert self.name is not None
        os.rename(self.name, new_location)
        self.name = None
    # enddef moveToPlace

    def __del__(self):
        self.safe_unlink()

def safeConfigWriter(filename, data, bprefix = ".", bsuffix = ".bak"):
    """This function allows to save a [configuration] in a safe manner.
    If the target file doesn't exist yet, it will be written using
    provided data. If it does exist then a temporary file will be first
    created, then existing file will be atomically moved to a backup
    location and the new file will be finally placed to the target
    location.

    If an error occurs at any point of the process, temporary files
    will be removed (well, at least an attempt will be made in most
    cases).

    If the backup file already exists, then it will NOT be overwriiten
    instead a random suffix will be appended to the backup name to form
    an unique one (similar to mktemp/mkstemp/NamedTemporaryFile).

    @param filename: a target configuration filename
    @type filename: a C{str} or an equivalent type
    @param data: an iterator supplying configuration file content (strings)
    @param bprefix: an optional prefix to form a backup filename ("." by default)
    @param bsuffix: an optional suffix to form a backup filename (".bak" by default)
    """
    newf = _NewConfigFile(filename, bprefix)

    # file(s) will be deleted upon an error (exception)
    # CORRECTION: they will be deleted upon garbage collecting
    # in case of an exception. This happens not immediately
    # after leaving this particular frame but rather after
    # the frame (stored in backtrace object) will be garbage
    # collected)
    newf.writelines(data)
    newf.close()

    if newf.name == filename:
        # everything's done, the new config file is on its place
        newf.name = None # prevent unlinking upon del
        return

    # need to create a backup file
    backname = os.path.join(os.path.dirname(filename),
                            bprefix + os.path.basename(filename) + bsuffix)
    backf = _NewConfigFile(backname, prefix='')
    # the file is created.
    # Actually we don't need it, just its name

    # some OSes doesn't want to move or unlink opened files
    backf.close()

    # here may be a race. simply die if it's the case,
    # things will be automatically cleaned up.
    backf_name = backf.name
    backf.safe_unlink()
    os.link(filename, backf_name)
    # at this point we have the new config in a temporary
    # location and the old config in both 'target' and
    # 'backup' locations, unmodified (including mtime,
    # permissions and possible EAs)

    # it's time to move the newly created config to its permanent
    # location
    newf.moveToPlace(filename)

    # voila! now take care the old backup file won't
    # be deleted upon object destruction (the new file
    # will not be deleted because it's moved to a permanent
    # location)
    backf.name = None
    _theLogger.info('backup copy of "%s" is saved as "%s"' % (filename, backf_name))
# endfef safeConfigWriter

class dummy_reversed(object):
    """A dumb implementation of "reversed" iterator over a
    fixed-length sequence
    """
    def __init__(self, seq):
        self._seq = seq
        self._cur = len(seq)

    def __iter__(self):
        return self

    def next(self):
        # FIXME: not thread safe
        self._cur -= 1
        if self._cur < 0:
            raise StopIteration
        return self._seq[self._cur]
    # enddef next
# endclass dummy_reversed

