# -*- coding: utf-8; mode: python -*-
"""This scripts analyzes current PAM configuration and
perform necessary changes to gracefully remove Plesk PAM
module from all configuration chains"""

import os.path

try: import logging
except ImportError: import compat.dummy_logging as logging

from traceback import format_exception, format_exception_only

# PAMConfig
from PAMConfig import ConfSet
from PAMConfig.analyzers import AuthConfAnalyzer, PasswordConfAnalyzer
from PAMConfig.modifiers import AuthConfModifier, PasswordConfModifier

from PAMConfig.defaults import KnownPasswordSetters

# common constants and functions
from pam_plesk_common import *

#### Functions which do most of the job :-)
def getAuthChanges(confSet, theModule):
    """collects changes which are to be done to remove the module
    from auth facility"""

    analyzer = AuthConfAnalyzer(theModule)
    # receive initial list of PAM services detected on this system
    (autonomous, commonDict) = analyzer.selectServices(confSet)
    logging.debug("initial services: autonomous = %s, commonDict = %s" % \
                  (autonomous, commonDict))

    # Now select services for which theModule is used
    autonomousCheckers = analyzer.selectImportant(autonomous, confSet,
                                                  [os.path.basename(theModule)], fullPath = False)
    # autonomousCheckers is a dict with keys being services names
    # and values being positions in the corresponding configuration
    # list where the change is planned
    commonCheckers = analyzer.selectImportant(commonDict.iterkeys(), confSet,
                                              [os.path.basename(theModule)], fullPath = False)
    # commonCheckers is organized similar to autonomousCheckers

    changes = []
    for (svc, pos) in autonomousCheckers.iteritems():
        style = analyzer.getRemoveStyle(svc, pos, confSet, [])
        changes.append((svc, pos, [], style))

    for (svc, pos) in commonCheckers.iteritems():
        style = analyzer.getRemoveStyle(svc, pos, confSet, commonDict[svc])
        changes.append((svc, pos, commonDict[svc], style))

    return changes

# enddef getAuthChanges

def getPasswordChanges(confSet, theModule):
    """collects changes which are to be done to remove the module
    from password facility"""
    analyzer = PasswordConfAnalyzer(theModule)
    # receive initial list of PAM services detected on this system
    (autonomous, commonDict) = analyzer.selectServices(confSet)
    logging.debug("initial services: autonomous = %s, commonDict = %s" % \
                  (autonomous, commonDict))

    # Now select services for which theModule is used
    autonomousCheckers = analyzer.selectImportant(autonomous, confSet,
                                                  [os.path.basename(theModule)], fullPath = False)
    # autonomousCheckers is a dict with keys being services names
    # and values being positions in the corresponding configuration
    # list where the change is planned
    commonCheckers = analyzer.selectImportant(commonDict.iterkeys(), confSet,
                                              [os.path.basename(theModule)], fullPath = False)
    # commonCheckers is organized similar to autonomousCheckers

    changes = []
    for (svc, pos) in autonomousCheckers.iteritems():
        style = analyzer.getRemoveStyle(svc, pos, confSet, KnownPasswordSetters)
        changes.append((svc, pos, style))

    for (svc, pos) in commonCheckers.iteritems():
        style = analyzer.getRemoveStyle(svc, pos, confSet, KnownPasswordSetters)
        changes.append((svc, pos, style))

    return changes
# enddef getPasswordChanges

def applyAuthChanges(changes, confSet, theModule):
    changes = sortChanges(changes)
    modifier = AuthConfModifier(theModule)
    modMethods = {
        'required':       modifier.removeRequired,
        'requiredSingle': modifier.removeRequiredSingle,
        'sufficient':     modifier.removeSufficient
        }

    for (svc, position, affected, style) in changes:
        try:
            meth = modMethods[style]
        except KeyError:
            raise NotImplementedError("do not know how to handle '%s' change style" % style)
        meth(confSet, svc, position, affected)
    # endfor
# enddef applyAuthChanges

def applyPasswordChanges(changes, confSet, theModule):
    changes = sortChanges(changes)
    modifier = PasswordConfModifier(theModule)
    modMethods = {
        'optional':           modifier.remove,
        'required':           modifier.remove,
        'requiredSingle':     modifier.removeRequiredSingle,
        }
    
    for (svc, position, style) in changes:
        try:
            meth = modMethods[style]
        except KeyError:
            raise NotImplementedError("do not know how to handle '%s' change style" % style)
        meth(confSet, svc, position)
    # endfor
# enddef applyPasswordChanges
    
def main():
    # be polite to our users and handle all exceptions specifically
    try:
        (options, args) = parseArgs()
        setupLoggers(options)
    except SystemExit:
        # in case we exited after usage()
        raise
    except:
        # Any error at this stage IS a programming error.
        msg = "a programming / runtime error"
        logging.exception(msg)
        exitmsg(3, msg)
    # endtry

    # Load given configuration set into memory
    try:
        logging.debug('Loading PAM configuration from (directory="%s", file="%s")' % \
                      (options.confDir, options.confFile))
        confSet = ConfSet(options.confFile, options.confDir)
    except EnvironmentError:
        # Likely a permission or filesystem consistency error
        (et, ev, tb) = sys.exc_info()
        exitmsg(2, ''.join(format_exception_only(et, ev)))
    except NotImplementedError:
        # report the error and attach a whole configuration (now
        # it can be structurally dumped from within confSet.
        # it's a better solution than a zipfile, I guess
        (et, ev, tb) = sys.exc_info()
        trace = 'S2:' + ''.join(format_exception(et, ev, tb))
        # TODO: it's good at this point to store entire set of PAM
        # configuration files
        logging.debug(trace)
        exitmsg(4, "some functionality isn't yet implemented")
    except:
        # likely a programming error. Basically the same as for
        # NotImplementedError
        (et, ev, tb) = sys.exc_info()
        trace = 'S2:' + ''.join(format_exception(et, ev, tb))
        # TODO: it's good at this point to store entire set of PAM
        # configuration files
        logging.debug(trace)
        exitmsg(3, "a programming/runtime error")
    # endtry
    logging.info("""\
PAM configuration loaded from (directory="%s", file="%s"). \
Configuration contains %d file(s)\
""" % (options.confDir, options.confFile, len(confSet.files)))

    # Now we need to calculate changes
    try:
        logging.debug("calculating changes for 'auth' stacks")
        changes = getAuthChanges(confSet, options.pleskModule)

        if changes:
            logging.info("calculated %d position(s) to change 'auth' stacks" % len(changes))
            applyAuthChanges(changes, confSet, options.pleskModule)
            logging.info("required 'auth' stack(s) modified (in memory)")

        logging.debug("calculating changes for 'password' stacks")
        changes = getPasswordChanges(confSet, options.pleskModule)
        if changes:
            logging.info("calculated %d position(s) to change 'password' stacks" % len(changes))
            applyPasswordChanges(changes, confSet, options.pleskModule)
            logging.info("required 'password' stack(s) modified (in memory)")
            
    except NotImplementedError:
        # report the error and attach a whole configuration (now
        # it can be structurally dumped from within confSet.
        # it's a better solution than a zipfile, I guess
        (et, ev, tb) = sys.exc_info()
        trace = 'S3:' + ''.join(format_exception(et, ev, tb))
        # TODO: it's good at this point to store entire set of PAM
        # configuration files. Now it can be done by dumping
        # confSet
        logging.debug(trace)
        exitmsg(4, "some functionality isn't yet implemented")
    except:
        # likely a programming error. Basically the same as for
        # NotImplementedError
        (et, ev, tb) = sys.exc_info()
        trace = 'S3:' + ''.join(format_exception(et, ev, tb))
        # TODO: it's good at this point to store entire set of PAM
        # configuration files. Now it can be done by dumping
        # confSet
        logging.debug(trace)
        exitmsg(3, "a programming/runtime error")
    # endtry

    # Time to write the changes back to the disk
    try:
        logging.debug("Ready to write changed stacks back to disk")
        confSet.sync()
        logging.info('PAM configuration saved (directory="%s", file="%s")' % \
                     (options.confDir, options.confFile))
    except EnvironmentError:
        # Likely a permission or filesystem consistency error
        (et, ev, tb) = sys.exc_info()
        exitmsg(2, ''.join(format_exception_only(et, ev)))
    except NotImplementedError, e:
        # report the error and attach a whole configuration (now
        # it can be structurally dumped from within confSet.
        # it's a better solution than a zipfile, I guess
        (et, ev, tb) = sys.exc_info()
        trace = 'S4:' + ''.join(format_exception(et, ev, tb))
        # TODO: it's good at this point to store entire set of PAM
        # configuration files. Now it can be done by dumping
        # confSet
        logging.debug(trace)
        exitmsg(4, "some functionality isn't yet implemented")
    except:
        # likely a programming error. Basically the same as for
        # NotImplementedError
        (et, ev, tb) = sys.exc_info()
        trace = 'S4:' + ''.join(format_exception(et, ev, tb))
        # TODO: it's good at this point to store entire set of PAM
        # configuration files. Now it can be done by dumping
        # confSet
        logging.debug(trace)
        exitmsg(3, "a programming/runtime error")
    # endtry

    exitmsg(0, "")
# enddef main

if __name__ == '__main__':
    main()
